@@ -1,6 +1,6 @@
 /*  Plzip - A parallel version of the lzip data compressor
     Copyright (C) 2009 Laszlo Ersek.
-    Copyright (C) 2009 Antonio Diaz Diaz.
+    Copyright (C) 2009, 2010 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -16,15 +16,17 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-struct Plzip_arg
-{
-  unsigned num_worker,
-      num_slot,
-      clidx;
-  int print_cctrs,
-      infd,
-      outfd;
-};
+struct Muxer_arg
+  {
+  int dictionary_size;
+  int match_len_limit;
+  int num_workers;
+  int num_slots;
+  int debug_level;
+  int infd;
+  int outfd;
+  };
 
-void *
-plzip_wrap(void *v_arg);
+void * muxer( void * arg );
+
+extern int verbosity;
